(*<scc.pas>
 * Library for SCC soundchip handling.
 * Thanks to BIFI's website at http://bifi.msxnet.org/msxnet//tech/scc
 * CopyLeft (c) since 1995 by PopolonY2k.
 *)

(**
  *
  * $Id: scc.pas 98 2015-08-21 01:28:40Z popolony2k $
  * $Author: popolony2k $
  * $Date: 2015-08-21 01:28:40 +0000 (Fri, 21 Aug 2015) $
  * $Revision: 98 $
  * $HeadURL: file:///svn/p/oldskooltech/code/msx/trunk/msxdos/pascal/scc.pas $
  *)

(*
 * This module depends on folowing include files (respect the order):
 * - types.pas;
 * - msxbios.pas;
 *)

(**
  * Important module constants.
  *)
Const
         ctSCCRegCounter  : Byte = $7F;  { SCC register counter }
         ctSCCPortCounter : Byte = $03;  { SCC port counter }


(**
  * Local module variables.
  *)
Var
      __aSCCAddress     : Array[0..3] Of Integer;
      __nSCCSlotNumber  : TSlotNumber;
      __nSCCPort,
      __nSCCRegister,
      __nSCCData        : Byte;


(**
  * Find the slot that SCC lives in.
  *)
Function FindSCC : TSlotNumber;
Var
        bResult         : Boolean;
        nCount          : Integer;
        nPrimarySlot,
        nSecondarySlot  : Byte;
        aEXPTBL         : Array[0..3] Of Byte Absolute $FCC1;
        nSlotNumber     : TSlotNumber;

Begin
  nPrimarySlot := 0;

  (* Search for SCC slot *)
  Repeat
    nSecondarySlot := 0;

    Repeat
      bResult := True;
      nSlotNumber := MakeSlotNumber( nPrimarySlot, nSecondarySlot );

      (*
       * Following the http://bifi.msxnet.org/msxnet/tech/scc we can find
       * that to activate the SCC on MSX is needed just to write $3F to
       * bank select register 3 (some place between memory address to $9000
       * to $97FF) to activate it.
       * After this, you can read and write at $9800 to $9FFF.
       *)
      WRSLT( nSlotNumber, $9000, $3F );

      (*
       * Check for a memory behavior specific of SCC soundchip.
       * The memory area from $9800 to $987F behaves as RAM, so we can write
       * something there and try to read the same content.
       *)
      For nCount := $9800 To $987F Do
      Begin
        WRSLT( nSlotNumber, nCount, $7F );
        bResult := bResult And ( RDSLT( nSlotNumber, nCount ) = $7F );
      End;

      If( bResult )  Then
      Begin
        (*
         * The memory area between $9880 to $98FF is write only, so if you
         * try to read it, it'll always return $FF.
         * WARNING: To the test below the range considered is just between
         * $9880 to $988E, because the $988F is a on/off switch to channels
         * 1 to 5, so if we test it they would be reseted and the SCC
         * channels won't play anymore.
         *)
        For nCount := $9880 To $988E Do
        Begin
          WRSLT( nSlotNumber, nCount, 1 );
          bResult := bResult And ( RDSLT( nSlotNumber, nCount ) = $FF );
        End;
      End;

      If( Not bResult )  Then
        nSecondarySlot := nSecondarySlot + 1;
    Until( bResult Or
           ( nSecondarySlot = ctMaxSecSlots ) Or
           ( aEXPTBL[nPrimarySlot] = 0 ) );

    If( Not bResult )  Then
      nPrimarySlot := nPrimarySlot + 1;
  Until( bResult Or ( nPrimarySlot = ctMaxSlots ) );

  If( Not bResult )  Then
    nSlotNumber := ctUnitializedSlot;

  (* Initialize the internal SCC array addresses *)
  __aSCCAddress[0] := $9800;
  __aSCCAddress[1] := $9880;
  __aSCCAddress[2] := $988A;
  __aSCCAddress[3] := $988F;

  FindSCC := nSlotNumber;
End;

(**
  * Write data to SCC sound chip using a direct mode by passing the
  * internal scc parameters __nSCCSlotNumber, __nSCCPort, __nSCCRegister,
  * __nSCCData.
  * @param __nSCCSlotNumber The @see TSlotNumber structure containing the
  * information of SCC card that will receive the information;
  * @param __nSCCPort The port that register are categorized;
  * @param __nSCCRegister The register index to write data;
  * @param __nSCCData The data to be written;
  *)
Procedure WriteSCCDirect{( __nSCCSlotNumber : TSlotNumber;
                           __nSCCPort, __nSCCRegister, __nSCCData : Byte ) };
Var
        nAddress  : Integer;
Begin
 {
 WRSLT( __nSCCSlotNumber,
        ( __aSCCAddress[__nSCCPort] + __nSCCRegister ),
        __nSCCData );
 }

 nAddress := ( __aSCCAddress[__nSCCPort] + __nSCCRegister );

 InLine( $3A/__nSCCSlotNumber/     { LD A,  (__nSCCSlotNumber) }
         $2A/nAddress/             { LD HL, (nAddress)         }
         $ED/$5B/__nSCCData/       { LD DE, (__nSCCData)       }
         $CD/$14/$00               { CALL WRSLT                }
       );
End;

(**
  * Write data to SCC sound chip.
  * @param nSlotNumber The @see TSlotNumber structure containing the
  * information of SCC card that will receive the information;
  * @param nPort The port that register are categorized;
  * @param nRegister The register index to write data;
  * @param nData The data to be written;
  *)
Procedure WriteSCC( nSlotNumber : TSlotNumber;
                    nPort, nRegister, nData : Byte );
Begin
  __nSCCSlotNumber := nSlotNumber;
  __nSCCPort       := nPort;
  __nSCCRegister   := nRegister;
  __nSCCData       := nData;
  WriteSCCDirect;
End;

(**
  * Reset the SCC soundchip.
  * @param nSlotNumber The @see TSlotNumber structure to access the soundchip.
  *)
Procedure ResetSCC( nSlotNumber : TSlotNumber );
Begin
  __nSCCSlotNumber := nSlotNumber;
  __nSCCData := 0;

  For __nSCCPort := 0 To ctSCCPortCounter Do
    For __nSCCRegister := 0 To ctSCCRegCounter Do
      WriteSCCDirect{( __nSCCSlotNumber,
                       __nSCCPort,
                       __nSCCRegister,
                       __nSCCData = 0 )};
End;
