(*<vgmplay.pas>
 * Library for VGM music handling.
 * Check for newer VGM format definition at following URL address below
 * http://www.smspower.org/Music/VGMFileFormat?from=Development.VGMFormat
 * CopyLeft (c) since 1995 by PopolonY2k.
 *)

(**
  *
  * $Id: vgmplay.pas 98 2015-08-21 01:28:40Z popolony2k $
  * $Author: popolony2k $
  * $Date: 2015-08-21 01:28:40 +0000 (Fri, 21 Aug 2015) $
  * $Revision: 98 $
  * $HeadURL: file:///svn/p/oldskooltech/code/msx/trunk/msxdos/pascal/vgmplay.pas $
  *)

(*
 * This module depends on folowing include files (respect the order):
 * - systypes.pas;
 * - system.pas;
 * - sleepb.pas;
 * - types.pas;
 * - msxdos.pas;
 * - msxdos2.pas;
 * - msxbios.pas;
 * - dos2file.pas;
 * - ay8910.pas
 * - scc.pas;
 * - opl4.pas;
 * - y8950.pas;
 * - ym2413.pas;
 * - ym2151.pas;
 * - vgmtypes.pas;
 * - sndchips.pas;
 *)

{$r-}

(**
  * VGM commands definitions.
  *)
Const
        ctWaitSamples           = $61;   { Wait for nnnn (16bit) samples }
        ctWait735Samples        = $62;   { Wait 735 samples }
        ctWait882Samples        = $63;   { Wait 882 samples }
        ctOverrideLength        = $64;   { Override length of $62/$63 }
        ctAY8910Write           = $A0;   { Write value dd to register aa }
        ctYM2413Write           = $51;   { Write value dd to register aa }
        ctY8950Write            = $5C;   { Write value dd to register aa }
        ctYM2151Write           = $54;   { Write value dd to register aa }
        ctK051649Write          = $D2;   { Port pp write val. dd to reg. aa }
        ctYMF278BWrite          = $D0;   { Port pp write val. dd to reg. aa }
        ctEndOfSoundData        = $66;   { End of sound data }


(**
  * Play the VGM file content previously opened by @see OpenVGM.
  * @param data The data that will be played;
  * @param chips The @see TSoundChips structure containing all information
  * about the known chips used by the library;
  *)
Procedure PlayVGM( Var data  : TVGMData;
                   Var chips : TSoundChips );

Var      nTimingDivisor,
         nCurrentStreamPos,
         nLoopOffsetStreamPos  : Integer;
         nCurrentVGMCommand    : Byte;
         bStopPlaying          : Boolean;

Label    __loop;     { For fast looping }


(* PlayVGM main entry point *)
Begin
  (* Memory based chips initialization *)
  __nSCCSlotNumber    := chips.nSCCSlotNumber;
  __nYM2151SlotNumber := chips.nYM2151SlotNumber;

  (*
   * Calculates the timing divisor that will be used
   * depending on clock passed by parameter.
   *)
  If( chips.intrTiming = Timing50Hz )  Then
    nTimingDivisor  := Round( ( 44100.0 / 50 ) / chips.nHostTickCount )
  Else
    nTimingDivisor  := Round( ( 44100.0 / 60 ) / chips.nHostTickCount );

  (*
   * Points the LoopOffset pointer correctly.
   * Copy just the 16Bit part of LoopOffset because the library won't
   * handle large files containing 32Bit amount of data.
   *)
  Move( data.header.nLoopOffset,
        nCurrentStreamPos,
        SizeOf( nCurrentStreamPos ) );

  (* Check if song has a loop *)
  If( nCurrentStreamPos <> 0 )  Then
    nLoopOffsetStreamPos := ( nCurrentStreamPos - data.nHeaderSize )
  Else
    nLoopOffsetStreamPos := -1;

  (* General internal initialization *)
  bStopPlaying      := False;
  nCurrentStreamPos := 0;

  DI;   { Disable interrupts }

  Repeat
    __loop:

    (*
     * Check for the ESC key to stop playing.
     *)
    Port[$AA] := 7;  { PPI keyboard matrix select }
    bStopPlaying := ( ( Port[$A9] And 4 ) = 0 );   { ESC key reading }

    If( bStopPlaying )  Then
      nCurrentVGMCommand := ctEndOfSoundData
    Else
      nCurrentVGMCommand := data.pVGMSongBuffer^[nCurrentStreamPos];

    nCurrentStreamPos := Succ( nCurrentStreamPos );

    Case nCurrentVGMCommand Of
      ctWaitSamples :    { Wait n samples - $61 nn nn }
        Begin
          Move( data.pVGMSongBuffer^[nCurrentStreamPos], __nIntrTickCount, 2 );

          __nIntrTickCount := ( __nIntrTickCount Div nTimingDivisor );
          SleepIntrTicksDirect{( __nIntrTickCount )};

          { Increases the pointer to the next stream position }
          nCurrentStreamPos := nCurrentStreamPos + 2;
          Goto __loop;
        End;

      ctWait735Samples,  { Wait 735 samples (50Hz) }
      ctWait882Samples : { Wait 882 samples (60Hz) }
        Begin
          __nIntrTickCount := chips.nHostTickCount;

          SleepIntrTicksDirect{( __nIntrTickCount )};
          Goto __loop;
        End;

      $70..$7F :         { Wait n+1 samples, n can range from 0 to 15 }
        Goto __loop;

      ctYM2151Write :     { YM2151 write dd to register aa }
        Begin
          __nYM2151AddrReg := data.pVGMSongBuffer^[nCurrentStreamPos];
          __nYM2151DataReg := data.pVGMSongBuffer^[nCurrentStreamPos + 1];

          WriteYM2151Direct{( __nYM2151SlotNumber,
                              __nYM2151AddrReg,
                              __nYM2151DataReg )};

          { Increases the pointer to the next stream position }
          nCurrentStreamPos := nCurrentStreamPos + 2;
          Goto __loop;
        End;

      ctK051649Write :   { K051649 port pp, write dd to register aa }
        Begin
          __nSCCPort     := data.pVGMSongBuffer^[nCurrentStreamPos];
          __nSCCRegister := data.pVGMSongBuffer^[nCurrentStreamPos + 1];
          __nSCCData     := data.pVGMSongBuffer^[nCurrentStreamPos + 2];

          WriteSCCDirect{( __nSCCSlotNumber,
                           __nSCCPort,
                           __nSCCRegister,
                           __nSCCData )};

          { Increases the pointer to the next stream position }
          nCurrentStreamPos := nCurrentStreamPos + 3;
          Goto __loop;
        End;

      ctYMF278BWrite :   { YMF278B port pp, write dd to register aa }
        Begin
          __nOPL4Port     := data.pVGMSongBuffer^[nCurrentStreamPos];
          __nOPL4Register := data.pVGMSongBuffer^[nCurrentStreamPos + 1];
          __nOPL4Data     := data.pVGMSongBuffer^[nCurrentStreamPos + 2];

          WriteOPL4FMDirect{( __nOPL4Port, __nOPL4Register, __nOPL4Data )};

          { Increases the pointer to the next stream position }
          nCurrentStreamPos := nCurrentStreamPos + 3;
          Goto __loop;
        End;

      ctY8950Write :     { Y8950 write dd to register aa }
        Begin
          __nY8950Register := data.pVGMSongBuffer^[nCurrentStreamPos];
          __nY8950Data     := data.pVGMSongBuffer^[nCurrentStreamPos + 1];

          WriteY8950Direct{( __nY8950Register, __nY8950Data )};

          { Increases the pointer to the next stream position }
          nCurrentStreamPos := nCurrentStreamPos + 2;
          Goto __loop;
        End;

      ctAY8910Write :    { AY8910, write dd to register aa }
        Begin
          __nAY8910Register := data.pVGMSongBuffer^[nCurrentStreamPos];
          __nAY8910Data     := data.pVGMSongBuffer^[nCurrentStreamPos + 1];

          WriteAY8910Direct{( __nAY8910Register, __nAY8910Data )};

          { Increases the pointer to the next stream position }
          nCurrentStreamPos := nCurrentStreamPos + 2;
          Goto __loop;
        End;

      ctYM2413Write :    { YM2413, write dd to register aa }
        Begin
          __nYM2413Index := data.pVGMSongBuffer^[nCurrentStreamPos];
          __nYM2413Data  := data.pVGMSongBuffer^[nCurrentStreamPos + 1];

          WriteYM2413Direct{( __nYM2413Index, __nYM2413Data )};

          { Increases the pointer to the next stream position }
          nCurrentStreamPos := nCurrentStreamPos + 2;
          Goto __loop;
        End;

      ctEndOfSoundData : { End Of sound data - Loop management }
        Begin
          If( Not bStopPlaying )  Then
            Begin
              nCurrentStreamPos := nLoopOffsetStreamPos;

              { End of streaming ?? }
              If( nLoopOffsetStreamPos = -1 )  Then
                bStopPlaying := True;
            End;
        End;

      { Reserved Operands and not supported chips }

      ctOverrideLength : { Override length of $62/$63 }
        Begin
          nCurrentStreamPos := nCurrentStreamPos + 3;
          Goto __loop;
        End;

      $80..$8F :         { YM2612 unsupported command. Pop!Art ignore it }
        Goto __loop;

      $30..$3F,          { One operand, reserved for future use }
      $4F,               { Game gear unsupported by Pop!Art }
      $50      :         { PSG (SN76489/SN76496) unsupported by Pop!Art }
        Begin
          nCurrentStreamPos := Succ( nCurrentStreamPos );
          Goto __loop;
        End;

      $40..$4E :         { Two operands (>=1.70), reserved for future use }
        Begin
          { Protocols until 1.60 is one operand }
          nCurrentStreamPos := Succ( nCurrentStreamPos );

          { For protocols since version 1.61 are two operands }
          If( data.header.nVersionNumber[0] >= $61 ) Then
            If( data.header.nVersionNumber[1] >= $1 ) Then
              nCurrentStreamPos := Succ( nCurrentStreamPos );

          Goto __loop;
        End;

      $A1..$AF,          { Two operands, reserved for future use }
      $B0..$BB,          { Known 1.70 protocol chips unsupported by Pop!Art }
      $BC..$BF,
      $52,
      $53,
      $55..$5B,
      $5D..$5F:          { Known 1.70 protocol chips unsupported by Pop!Art }
        Begin
          nCurrentStreamPos := nCurrentStreamPos + 2;
          Goto __loop;
        End;

      $C0..$C4 :         { Known 1.70 protocol chips unsupported by Pop!Art }
        Begin
          nCurrentStreamPos := nCurrentStreamPos + 6;
          Goto __loop;
        End;

      $C5..$CF,
      $D5..$DF :         { Three operands, reserved for future use }
        Begin
          nCurrentStreamPos := nCurrentStreamPos + 3;
          Goto __loop;
        End;

      $E1..$FF :         { Four operands, reserved for future use }
        Begin
          nCurrentStreamPos := nCurrentStreamPos + 4;
          Goto __loop;
        End;
    End;
  Until( bStopPlaying );

  EI;  { Enable interrupts }
End;
